@@ -11,7 +11,10 @@ use pyo3::{
     types::{PyComplex, PyInt},
     IntoPy, Py, PyObject, PyResult, Python, ToPyObject,
 };
-use qcs::qpu::api::{ConnectionStrategy, ExecutionOptions, ExecutionOptionsBuilder};
+use qcs::qpu::api::{
+    ApiExecutionOptions, ApiExecutionOptionsBuilder, ConnectionStrategy, ExecutionOptions,
+    ExecutionOptionsBuilder,
+};
 use qcs_api_client_grpc::models::controller::{
     data_value, readout_values, ControllerJobExecutionResult,
 };
@@ -33,15 +36,21 @@ create_init_submodule! {
         ExecutionResults,
         PyConnectionStrategy,
         PyExecutionOptions,
-        PyExecutionOptionsBuilder
+        PyExecutionOptionsBuilder,
+        PyApiExecutionOptions,
+        PyApiExecutionOptionsBuilder
     ],
     errors: [
         SubmissionError,
-        RetrieveResultsError
+        QpuApiError
     ],
     funcs: [
         py_submit,
         py_submit_async,
+        py_cancel_job,
+        py_cancel_job_async,
+        py_cancel_jobs,
+        py_cancel_jobs_async,
         py_retrieve_results,
         py_retrieve_results_async
     ],
@@ -102,13 +111,8 @@ py_function_sync_async! {
     }
 }
 
-wrap_error!(RustRetrieveResultsError(qcs::qpu::api::QpuApiError));
-py_wrap_error!(
-    runner,
-    RustRetrieveResultsError,
-    RetrieveResultsError,
-    PyRuntimeError
-);
+wrap_error!(RustQpuApiError(qcs::qpu::api::QpuApiError));
+py_wrap_error!(runner, RustQpuApiError, QpuApiError, PyRuntimeError);
 
 /// Variants of data vectors within a single ExecutionResult.
 #[derive(Clone, Debug)]
@@ -271,6 +275,54 @@ impl ExecutionResults {
     }
 }
 
+py_function_sync_async! {
+    #[pyfunction]
+    #[pyo3(signature = (job_ids, quantum_processor_id = None, client = None, execution_options = None))]
+    async fn cancel_jobs(
+        job_ids: Vec<String>,
+        quantum_processor_id: Option<String>,
+        client: Option<PyQcsClient>,
+        execution_options: Option<PyExecutionOptions>,
+    ) -> PyResult<()> {
+        let client = PyQcsClient::get_or_create_client(client).await;
+
+        qcs::qpu::api::cancel_jobs(
+            job_ids.into_iter().map(|id| id.into()).collect(),
+            quantum_processor_id.as_deref(),
+            &client,
+            execution_options.unwrap_or_default().as_inner()
+        )
+        .await
+        .map_err(RustQpuApiError::from).map_err(RustQpuApiError::to_py_err)?;
+
+        Ok(())
+    }
+}
+
+py_function_sync_async! {
+    #[pyfunction]
+    #[pyo3(signature = (job_id, quantum_processor_id = None, client = None, execution_options = None))]
+    async fn cancel_job(
+        job_id: String,
+        quantum_processor_id: Option<String>,
+        client: Option<PyQcsClient>,
+        execution_options: Option<PyExecutionOptions>,
+    ) -> PyResult<()> {
+        let client = PyQcsClient::get_or_create_client(client).await;
+
+        qcs::qpu::api::cancel_job(
+            job_id.into(),
+            quantum_processor_id.as_deref(),
+            &client,
+            execution_options.unwrap_or_default().as_inner()
+        )
+        .await
+        .map_err(RustQpuApiError::from).map_err(RustQpuApiError::to_py_err)?;
+
+        Ok(())
+    }
+}
+
 py_function_sync_async! {
     #[pyfunction]
     #[pyo3(signature = (job_id, quantum_processor_id = None, client = None, execution_options = None))]
@@ -284,8 +336,8 @@ py_function_sync_async! {
 
         let results = qcs::qpu::api::retrieve_results(job_id.into(), quantum_processor_id.as_deref(), &client, execution_options.unwrap_or_default().as_inner())
             .await
-            .map_err(RustRetrieveResultsError::from)
-            .map_err(RustRetrieveResultsError::to_py_err)?;
+            .map_err(RustQpuApiError::from)
+            .map_err(RustQpuApiError::to_py_err)?;
 
         Python::with_gil(|py| {
             ExecutionResults::from_controller_job_execution_result(py, results)
@@ -300,6 +352,13 @@ py_wrap_type! {
 impl_repr!(PyExecutionOptions);
 impl_as_mut_for_wrapper!(PyExecutionOptions);
 
+py_wrap_type! {
+    #[derive(Debug, Default)]
+    PyApiExecutionOptions(ApiExecutionOptions) as "APIExecutionOptions"
+}
+impl_repr!(PyApiExecutionOptions);
+impl_as_mut_for_wrapper!(PyApiExecutionOptions);
+
 #[pymethods]
 impl PyExecutionOptions {
     #[staticmethod]
@@ -324,6 +383,13 @@ impl PyExecutionOptions {
             .map(|timeout| timeout.as_secs_f64())
     }
 
+    #[getter]
+    fn api_options(&self) -> Option<PyApiExecutionOptions> {
+        self.as_inner()
+            .api_options()
+            .map(|x| PyApiExecutionOptions(x.clone().into()))
+    }
+
     fn __richcmp__(&self, py: Python<'_>, other: &Self, op: CompareOp) -> PyObject {
         match op {
             CompareOp::Eq => (self.as_inner() == other.as_inner()).into_py(py),
@@ -332,6 +398,24 @@ impl PyExecutionOptions {
     }
 }
 
+#[pymethods]
+impl PyApiExecutionOptions {
+    #[staticmethod]
+    fn default() -> Self {
+        Self::from(ApiExecutionOptions::default())
+    }
+
+    #[staticmethod]
+    fn builder() -> PyApiExecutionOptionsBuilder {
+        PyApiExecutionOptionsBuilder::default()
+    }
+
+    #[getter]
+    fn bypass_settings_protection(&self) -> bool {
+        self.as_inner().bypass_settings_protection()
+    }
+}
+
 py_wrap_type! {
     PyExecutionOptionsBuilder(ExecutionOptionsBuilder) as "ExecutionOptionsBuilder"
 }
@@ -361,6 +445,16 @@ impl PyExecutionOptionsBuilder {
         );
     }
 
+    #[setter]
+    fn api_options(&mut self, api_options: Option<PyApiExecutionOptions>) {
+        *self = Self::from(
+            self.as_inner()
+                .clone()
+                .api_options(api_options.map(|x| x.into_inner().into()))
+                .clone(),
+        );
+    }
+
     #[setter]
     fn timeout_seconds(&mut self, timeout_seconds: Option<f64>) {
         let timeout = timeout_seconds.map(Duration::from_secs_f64);
@@ -376,6 +470,44 @@ impl PyExecutionOptionsBuilder {
     }
 }
 
+py_wrap_type! {
+    PyApiExecutionOptionsBuilder(ApiExecutionOptionsBuilder) as "APIExecutionOptionsBuilder"
+}
+
+#[pymethods]
+impl PyApiExecutionOptionsBuilder {
+    #[new]
+    fn new() -> Self {
+        Self::default()
+    }
+
+    #[staticmethod]
+    fn default() -> Self {
+        Self::from(ApiExecutionOptionsBuilder::default())
+    }
+
+    #[setter]
+    fn bypass_settings_protection(&mut self, bypass_settings_protection: bool) {
+        // `derive_builder::Builder` doesn't implement AsMut, meaning we can't use `PyWrapperMut`,
+        // which forces us into this awkward clone.
+
+        *self = Self::from(
+            self.as_inner()
+                .clone()
+                .bypass_settings_protection(bypass_settings_protection)
+                .clone(),
+        );
+    }
+
+    fn build(&self) -> PyResult<PyApiExecutionOptions> {
+        Ok(PyApiExecutionOptions::from(
+            self.as_inner()
+                .build()
+                .map_err(|err| PyValueError::new_err(err.to_string()))?,
+        ))
+    }
+}
+
 py_wrap_type! {
     #[derive(Default)]
     PyConnectionStrategy(ConnectionStrategy) as "ConnectionStrategy"
